/*
 * oifits.c
 *
 * Copyright (C) 2003-2017 Fabien Baron, Hrobjartur Thorsteinsson, John Young
 *
 * This file is part of BSMEM:
 * http://www.mrao.cam.ac.uk/research/optical-interferometry/bsmem-software/
 *
 * BSMEM is licensed under a Creative Commons
 * Attribution-NonCommercial-NoDerivatives 4.0 International License.
 * See LICENSE.txt for more details.
 */

/* Oifits routines.
* This package uses the Oifits Exchange routines by John Young to view,
* select and extract oi data.
*/
#include "bsmem.h"
#include "oifile.h"
#include "oifilter.h"

#include <stdlib.h>
#include <math.h>
#include <string.h>

#define uv_threshold 5.0e-5


int compare_uv(oi_uv uv, oi_uv withuv, float thresh)
{
	float sqthresh;
	float pcompu,pcompv;
	float mcompu,mcompv;


	sqthresh = thresh*thresh;

	pcompu = 2.0*(uv.u-withuv.u)*(uv.u-withuv.u)/(uv.u*uv.u+withuv.u*withuv.u);
	pcompv = 2.0*(uv.v-withuv.v)*(uv.v-withuv.v)/(uv.v*uv.v+withuv.v*withuv.v);
	mcompu = 2.0*(uv.u+withuv.u)*(uv.u+withuv.u)/(uv.u*uv.u+withuv.u*withuv.u);
	mcompv = 2.0*(uv.v+withuv.v)*(uv.v+withuv.v)/(uv.v*uv.v+withuv.v*withuv.v);

	/* To handle zeros */
	if((uv.u-withuv.u) == 0.0)
	{
		pcompu = 0.0;
	}
	if((uv.v-withuv.v) == 0.0)
	{
		pcompv = 0.0;
	}
	if((uv.u+withuv.u) == 0.0)
	{
		mcompu = 0.0;
	}
	if((uv.v+withuv.v) == 0.0)
	{
		mcompv = 0.0;
	}


	/* If uv is same as withuv then return 1 */
	if((pcompu<sqthresh)&&(pcompv<sqthresh))
	{
		return 1;
	}
	/* If uv is same as withuv but conjugated then return -1 */
	if((mcompu<sqthresh)&&(mcompv<sqthresh))
	{
		return -1;
	}

	return 0;
}

/**
 * Read OIFITS data for user-specified target and wavelength range.
 *
 * Conjugates all uv coords into the +ve u half plane.
 *
 * Assumes the following attributes of param have already
 * been set: datafile, minband, maxband, target_id, fluxerr
 */
int get_oi_fits_data(const RECONST_PARAMETERS *param, oi_data *data,
                     int *status)
{
  oi_fits all, filtered;
  oi_filter_spec filter;
  GList *link;
  oi_vis2 *pVis2;
  oi_t3 *pT3;
  oi_wavelength *pWave;
  int i, k, l, npve;
  long numvis2, numt3;
  oi_uv uv1, uv2, uv3;

  if (*status) return *status;

  /* Read OIFITS file */
  read_oi_fits(param->datafile, &all, status);
  if (*status) return *status;

  /* Filter data by target ID and wavelength */
  init_oi_filter(&filter);
  filter.target_id = param->target_id;
  filter.wave_range[0] = param->minband/billion;
  filter.wave_range[1] = param->maxband/billion;
  apply_oi_filter(&all, &filter, &filtered);

  /* Allocate memory */
  count_oi_fits_data(&filtered, NULL, &numvis2, &numt3);
  data->pow = malloc((1 + numvis2) * sizeof(data->pow[0]));
  data->powerr = malloc((1 + numvis2) * sizeof(data->powerr[0]));
  data->bisamp = malloc(numt3 * sizeof(data->bisamp[0]));
  data->bisamperr = malloc(numt3 * sizeof(data->bisamperr[0]));
  data->bisphs = malloc(numt3 * sizeof(data->bisphs[0]));
  data->bisphserr = malloc(numt3 * sizeof(data->bisphserr[0]));
  data->bsref = malloc(numt3 * sizeof(oi_bsref));
  /* we allocate as much space for UV as possible initially and then
   * reallocate in the end */
  data->uv = malloc((1 + numvis2 + 3*numt3) * sizeof(oi_uv));

  /* Add total "zero-baseline" flux component for normalization */
  data->npow = 1;
  data->nuv  = 1;
  data->pow[0] = 1.0;
  data->powerr[0] = param->fluxerr;
  data->uv[0].u = 0.;
  data->uv[0].v = 0.;

  /* Read powerspectra */
  link = filtered.vis2List;
  while (link != NULL)
  {
    pVis2 = (oi_vis2 *)link->data;
    pWave = oi_fits_lookup_wavelength(&filtered, pVis2->insname);
    g_assert(pWave != NULL); //:TODO:
    for (i=0; i<pVis2->numrec; i++) {
      for (k=0; k<pVis2->nwave; k++) {
        if (!pVis2->record[i].flag[k]) {
          data->pow[data->npow] = (float)pVis2->record[i].vis2data[k];
          data->powerr[data->npow] = (float)pVis2->record[i].vis2err[k];
          data->uv[data->nuv].u = (float)(pVis2->record[i].ucoord
                                          / pWave->eff_wave[k]);
          data->uv[data->nuv].v = (float)(pVis2->record[i].vcoord
                                          / pWave->eff_wave[k]);
          /* flip into +u half plane */
          if(data->uv[data->nuv].u < 0.0) {
            data->uv[data->nuv].u = -data->uv[data->nuv].u;
            data->uv[data->nuv].v = -data->uv[data->nuv].v;
          }
          data->npow++;
          g_assert_cmpint(data->npow, <=, (1 + numvis2));
          data->nuv++;
          g_assert_cmpint(data->nuv, <=,
                          (1 + numvis2 + 3*numt3));
        }
      }
    }
    link = link->next;
  }
  
  /* Read bispectra */
  link = filtered.t3List;
  data->nbis = 0;
  while (link != NULL)
  {
    pT3 = (oi_t3 * )link->data;
    pWave = oi_fits_lookup_wavelength(&filtered, pT3->insname);
    g_assert(pWave != NULL); //:TODO:
    for (i=0; i<pT3->numrec; i++) {
      for (k=0; k<pT3->nwave; k++) {
        if (!pT3->record[i].flag[k]) {

          /* Trick to use closure data without available bis amplitudes */
          if(isnan(pT3->record[i].t3amp[k])) {
            data->bisamp[data->nbis] = 1.0;
            data->bisamperr[data->nbis] = infinity ;
          } else {
            data->bisamp[data->nbis] = (float)(pT3->record[i].t3amp[k]);
            data->bisamperr[data->nbis] = (float)(pT3->record[i].t3amperr[k]);
          }

          data->bisphs[data->nbis] = (float)(pT3->record[i].t3phi[k]);
          data->bisphserr[data->nbis] = (float)(pT3->record[i].t3phierr[k]);

          /* Read UV coords and check if they exist. If do not exist -> update UV.
           * Set the bsref.
           */
          uv1.u = (float)(pT3->record[i].u1coord / pWave->eff_wave[k]);
          uv1.v = (float)(pT3->record[i].v1coord / pWave->eff_wave[k]);
          uv2.u = (float)(pT3->record[i].u2coord / pWave->eff_wave[k]);
          uv2.v = (float)(pT3->record[i].v2coord / pWave->eff_wave[k]);
          uv3.u = -(uv1.u + uv2.u);
          uv3.v = -(uv1.v + uv2.v);

          /* Check if UV1, UV2, UV3 exist... */ //:TODO: move to sub-function

          /* ...uv1 */
          data->bsref[data->nbis].ab.uvpnt = -1;
          for (l=0; l<data->nuv; l++) {
            if ((npve = compare_uv(uv1, data->uv[l], uv_threshold))) {
              data->bsref[data->nbis].ab.uvpnt = l;
              data->bsref[data->nbis].ab.sign = npve;
              break; /* so that first match is referenced */
            }
          }
          if (data->bsref[data->nbis].ab.uvpnt == -1) {
            /* create new uv point */
            data->uv[data->nuv].u = uv1.u;
            data->uv[data->nuv].v = uv1.v;
            data->bsref[data->nbis].ab.uvpnt = l;
            data->bsref[data->nbis].ab.sign = 1;
          
            /* conjugate if u -ve */
            if (data->uv[data->nuv].u < 0.0) {
              data->uv[data->nuv].u = -uv1.u;
              data->uv[data->nuv].v = -uv1.v;
              data->bsref[data->nbis].ab.sign = -1;
            }
            data->nuv++;
            g_assert_cmpint(data->nuv, <=, (1 + numvis2 + 3*numt3));
          }
        
          /* ...uv2 */
          data->bsref[data->nbis].bc.uvpnt = -1;
          for (l=0; l<data->nuv; l++) {
            if ((npve = compare_uv(uv2, data->uv[l], uv_threshold))) {
              data->bsref[data->nbis].bc.uvpnt = l;
              data->bsref[data->nbis].bc.sign = npve;
              break;
            }
          }
          if (data->bsref[data->nbis].bc.uvpnt == -1) {
            /* create new uv point */
            data->uv[data->nuv].u = uv2.u;
            data->uv[data->nuv].v = uv2.v;
            data->bsref[data->nbis].bc.uvpnt = l;
            data->bsref[data->nbis].bc.sign = 1;

            /* conjugate if u -ve */
            if (data->uv[data->nuv].u < 0.0) {
              data->uv[data->nuv].u = -uv2.u;
              data->uv[data->nuv].v = -uv2.v;
              data->bsref[data->nbis].bc.sign = -1;
            }
            data->nuv++;
            g_assert_cmpint(data->nuv, <=, (1 + numvis2 + 3*numt3));
          }

          /* ...uv3 = (-uv2-uv2) */
          data->bsref[data->nbis].ca.uvpnt = -1;
          for (l=0; l<data->nuv; l++) {
            if ((npve = compare_uv(uv3, data->uv[l], uv_threshold))) {
              data->bsref[data->nbis].ca.uvpnt = l;
              data->bsref[data->nbis].ca.sign = npve;
              break;
            }
          }
          if (data->bsref[data->nbis].ca.uvpnt == -1)  {
            /* create new uv point */
            data->uv[data->nuv].u = uv3.u;
            data->uv[data->nuv].v = uv3.v;
            data->bsref[data->nbis].ca.uvpnt = l;
            data->bsref[data->nbis].ca.sign = 1;

            /* conjugate if u -ve */
            if (data->uv[data->nuv].u < 0.0) {
              data->uv[data->nuv].u = -uv3.u;
              data->uv[data->nuv].v = -uv3.v;
              data->bsref[data->nbis].ca.sign = -1;
            }
            data->nuv++;
            g_assert_cmpint(data->nuv, <=, (1 + numvis2 + 3*numt3));
          }

          data->nbis++;
          g_assert_cmpint(data->nbis, <=, numt3);
        }
      }
    }
    link = link->next;
  }

  g_assert_cmpint(data->npow, ==, (1 + numvis2));
  g_assert_cmpint(data->nbis, ==, numt3);
  data->uv = realloc(data->uv, data->nuv*sizeof(oi_uv));
  free_oi_fits(&all);
  free_oi_fits(&filtered);
  return *status;
}


//:TODO: pass in pointer to oi_fits to avoid reading data twice
/**
 * Get target and wavelength range to select data from OIFITS file.
 *
 * Uses the following attributes of param: datafile, minband, maxband
 *
 * Sets the following attributes of param: target_id, minband,
 * maxband, targetname
 */
int get_oi_fits_selection(RECONST_PARAMETERS *param, int *status)
{
  char blank[FLEN_VALUE], input[FLEN_VALUE];
  oi_fits all;
  oi_filter_spec filter;
  oi_wavelength *pWave;
  int i, k, id, numChannels, numConvert;
  GList *link;

  if (*status) return *status;

  for (k=0; k<FLEN_VALUE-1; k++)
    blank[k] = ' ';
  blank[FLEN_VALUE-1] = '\0';

  read_oi_fits(param->datafile, &all, status);
  if (*status) return *status;

  /* List targets */
  for(i=0; i<all.targets.ntarget; i++) {
    printf("\nTarget id/name:\t\t%d/%s\n",
           all.targets.targ[i].target_id, all.targets.targ[i].target);
  }

  /* Select target */
  if (all.targets.ntarget == 1) {
    printf("Auto selecting the only target \"%s\".\n",
           all.targets.targ[0].target);
    param->target_id = all.targets.targ[0].target_id;
    strncpy(param->targetname, all.targets.targ[0].target, FLEN_VALUE);
  } else {
    param->target_id = -1;
    while (param->target_id == -1) {
      printf("\nSELECT AN ID: ");
      scanf("%d", &id);
      for (i=0; i<all.targets.ntarget; i++) {
        if (all.targets.targ[i].target_id == id) {
          param->target_id = id;
          strncpy(param->targetname, all.targets.targ[i].target, FLEN_VALUE);
          break;
        }
      }
    }
  }

  /* PRINT AVAILABLE DATA ON SELECTED TARGET */
  print_oi_fits_summary(&all); //:TODO: only for selected target

  /* List wavebands by instrument */
  link = all.wavelengthList;
  numChannels = 0;
  i = 0;
  while (link != NULL) {
    pWave = (oi_wavelength *)link->data;
    numChannels += pWave->nwave;
    for (k=0; k<pWave->nwave; k++) {
      if (k == 0)
        printf("%-6d\t%-25.20s", i, pWave->insname);
      else
        printf("%-6.6s\t%-25.20s", blank, blank);
      printf("%-3.3d_%-14.3d\t%.0f/%.0f\n",
             i, k, pWave->eff_wave[k]*billion, pWave->eff_band[k]*billion);
    }
    link = link->next;
    ++i;
  }

  /* Select wavelength range */
  if (param->minband < 0. || param->maxband <= param->minband) {
    /* waveband not set (well) externally */
    if (numChannels > 1) {
      do {
        printf("Select a wavelength range (default value = 1 50000) :");
        fgets(input, FLEN_VALUE, stdin);
        numConvert = sscanf(input, "%f %f", &param->minband, &param->maxband);
        if (numConvert != 2)
          printf("Invalid band selection!\n");
      } while (numConvert != 2);
    } else {
      printf("Only one spectral channel. ");
      printf("Automatic selection of the full channel\n");
      param->minband = 1.;
      param->maxband = 50000.;
    }
  }

  free_oi_fits(&all);
  return *status;
}

void free_oi_data(oi_data *data)
{
	free(data->pow);
	free(data->powerr);
	free(data->bisamp);
	free(data->bisamperr);
	free(data->bisphs);
	free(data->bisphserr);
	free(data->uv);
	free(data->bsref);
}

float bsuv_coverage_quality(oi_bsref *bsref, int nbs, oi_uv *uv, int nuv)
{
	int i,k;
	float umax  = 0.0;
	float dumax = 0.0;
	float ds,ds1,dl,init;
	oi_uv u1, u2, du;
	float tmp, Q;
	oi_uv *UV;
	int nbsuv;

	nbsuv = 6*nbs;

	for(i=0; i<nuv; i++)
	{
		tmp = (float)sqrt((float)(uv[i].u*uv[i].u+uv[i].v*uv[i].v));
		if(tmp>umax)
		{
			umax = tmp;
		}
	}

	UV = (oi_uv*)malloc(nbsuv*sizeof(oi_uv));
	for(i=0; i<nbs; i++)
	{
		UV[6*i  ].u = uv[bsref[i].ab.uvpnt].u;
		UV[6*i  ].v = uv[bsref[i].ab.uvpnt].v;
		UV[6*i+1].u = -UV[6*i  ].u;
		UV[6*i+1].v = -UV[6*i  ].v;

		UV[6*i+2].u = uv[bsref[i].bc.uvpnt].u;
		UV[6*i+2].v = uv[bsref[i].bc.uvpnt].v;
		UV[6*i+3].u = -UV[6*i+2].u;
		UV[6*i+3].v = -UV[6*i+2].v;

		UV[6*i+4].u = uv[bsref[i].ca.uvpnt].u;
		UV[6*i+4].v = uv[bsref[i].ca.uvpnt].v;
		UV[6*i+5].u = -UV[6*i+4].u;
		UV[6*i+5].v = -UV[6*i+4].v;
	}

	init = (UV[0].u-UV[1].u)*(UV[0].u-UV[1].u)+(UV[0].v-UV[1].v)*(UV[0].v-UV[1].v);
	init = sqrt(init);
	ds  = init;
	dl  = 0.0;
	ds1 = init;

	for(i=0; i<(nbsuv-1); i++)
	{
		for(k=i+1; k<nbsuv; k++)
		{
			u1 = UV[i];
			u2 = UV[k];

			if(compare_uv(u1, u2, uv_threshold)==0)
			{
				du.u = u2.u - u1.u;
				du.v = u2.v - u1.v;

				tmp = sqrt(du.u*du.u+du.v*du.v);
				if(tmp>dumax)dumax = tmp;

				if(k == (i+1))ds1 = tmp;
				if(tmp<ds1)ds1 = tmp;
			}
		}
		if(ds1<ds)ds = ds1;
		if(ds1>dl)dl = ds1;
	}

	printf("dumax, umax, ds, dl = %e %e %e %e\n",dumax,umax,ds,dl);

	Q = dumax*ds/(2.0*umax*dl);

	return Q;
}

int count_redundant_bsuv(oi_bsref *bsref, int nbs)
{
	int i,k;
	int c = 0;
	int fab,fbc,fca;

	for(i=0; i<nbs; i++)
	{
		fab = 0;
		fbc = 0;
		fca = 0;
		for(k=0; k<nbs; k++)
		{
			if( (bsref[i].ab.uvpnt == bsref[k].ab.uvpnt) ||
				(bsref[i].ab.uvpnt == bsref[k].bc.uvpnt) ||
			(bsref[i].ab.uvpnt == bsref[k].ca.uvpnt)    )fab++;

			if( (bsref[i].bc.uvpnt == bsref[k].ab.uvpnt) ||
				(bsref[i].bc.uvpnt == bsref[k].bc.uvpnt) ||
			(bsref[i].bc.uvpnt == bsref[k].ca.uvpnt)    )fbc++;

			if( (bsref[i].ca.uvpnt == bsref[k].ab.uvpnt) ||
				(bsref[i].ca.uvpnt == bsref[k].bc.uvpnt) ||
			(bsref[i].ca.uvpnt == bsref[k].ca.uvpnt)    )fca++;
		}

		if(fab>1)c++;
		if(fbc>1)c++;
		if(fca>1)c++;
	}

	return c;
}

int read_fits_image(char* fname, float* img, int *n, float* xyint, char* source, char* datafile, int* status)
{
	fitsfile *fptr;       /* pointer to the FITS file, defined in fitsio.h */
	int  nfound, anynull;
	long naxes[2], fpixel, npixels;
	float nullval;

	if (*status==0)fits_open_file(&fptr, fname, READONLY, status);

	// MODIFY so that if keys do not exist, don't crash ^^

	if (*status==0)fits_read_keys_lng(fptr, "NAXIS", 1, 2, naxes, &nfound, status);
	//  if (*status==0)fits_read_key_str(fptr, "DATAFILE", datafile, comment, status);
	//  if (*status==0)fits_read_key_str(fptr, "TARGET", target, comment, status);
	//  if (*status==0)fits_read_key_flt(fptr, "PIXELATION", xyint, comment, status);
	npixels  = naxes[0] * naxes[1];         /* number of pixels in the image */
	fpixel   = 1;
	nullval  = 0;                /* don't check for null values in the image */

	//add here a compatibility check of xyint if imported as a model


	if(naxes[0] != naxes[1])
	{
		printf("Image dimension are not square.\n");
		if(*status==0)fits_close_file(fptr, status);
		return *status;
	}
	*n = naxes[0];
	/* Allocate enough memory outside of this routine */
	if(*status==0)fits_read_img(fptr, TFLOAT, fpixel, npixels, &nullval, img, &anynull, status);

	// renormalize the model !!!


	// reset the display by dispdev correctly


	if(*status==0)fits_close_file(fptr, status);

	return *status;
}
