/*
 * utest_oifits.c
 *
 * Copyright (C) 2014-2017 John Young
 *
 * This file is part of BSMEM:
 * http://www.mrao.cam.ac.uk/research/optical-interferometry/bsmem-software/
 *
 * BSMEM is licensed under a Creative Commons
 * Attribution-NonCommercial-NoDerivatives 4.0 International License.
 * See LICENSE.txt for more details.
 */

/*
 * Unit tests of OIFITS parsing.
 */

#include "bsmem.h"
#include <glib.h>

#define MAX_WL (50000.)
#define DIR "test/OIFITS1/"

typedef struct {
  const char *datafile;
  float minband;
  float maxband;
  int numtarget;
  int numvis2;
  int numt3;
  int nuv;  /**< number of unique uv points including one at zero frequency */
} TestCase;

typedef struct {
  RECONST_PARAMETERS param;
  oi_data data;
  int status;
} TestFixture;


//:TODO: commented out cases result in user prompt
static const TestCase cases[] = {
  //{DIR "Contest12_H.oifits",               0., MAX_WL, 1,  600,  800,  601},
  {DIR "Contest1_JHK.oifits",              0., MAX_WL, 1, 1800, 2400, 1801},
  {DIR "Mystery-Low_HK_1949nm.oifits",     0., MAX_WL, 1,  180,   61,  184},
  {DIR "Mystery-Low_HK_flagt3.oifits",     0., MAX_WL, 1, 3626, 1151, 3631},
  {DIR "Mystery-Low_HK_flagvis2.oifits",   0., MAX_WL, 1,  994, 1220, 3661},
  {DIR "Mystery-Low_HK.oifits",            0., MAX_WL, 1, 3660, 1220, 3661},
  {DIR "Mystery-Low_HK_split.oifits",      0., MAX_WL, 1, 3626, 1220, 3661},
  {DIR "Mystery-Low_HK_t3.oifits",         0., MAX_WL, 1,    0, 1220, 3661},
  {DIR "Mystery-Low_HK_vis2.oifits",       0., MAX_WL, 1, 3626,    0, 3627},
  {DIR "Mystery-Low_HK_vis2_t3phi.oifits", 0., MAX_WL, 1, 3626, 1220, 3661},
  //{DIR "Mystery-Med_H-AmberVISPHI_vis.oifits", 0., MAX_WL, 1,    0,    0,    1},
  {DIR "Contest1_JHK.oifits",           1520.,  1580., 1,  150,  200,  151},
  {DIR "Mystery-Low_HK.oifits",         1600.,  1800., 1,  732,  244,  733}
};

static const int numCases = sizeof(cases)/sizeof(cases[0]);


static gboolean ignoreRemoved(const char *logDomain, GLogLevelFlags logLevel,
                              const char *message, gpointer userData)
{
  return (!g_str_has_suffix(message, "removed from filter output"));
}

static void setup_fixture(TestFixture *fix, gconstpointer userData)
{
  fix->param.datafile[0] = '\0';
  fix->param.minband = 0.;
  fix->param.maxband = MAX_WL;
  fix->param.fluxerr = 1e-3;
  fix->status = 0;
}

static void teardown_fixture(TestFixture *fix, gconstpointer userData)
{
}

static void test_get_selection(TestFixture *fix, gconstpointer userData)
{
  int i, ret;

  for (i = 0; i < numCases; i++) {
    g_message("Scanning %s", cases[i].datafile);
    g_strlcpy(fix->param.datafile, cases[i].datafile, FLEN_FILENAME);
    fix->param.minband = cases[i].minband;
    fix->param.maxband = cases[i].maxband;
    ret = get_oi_fits_selection(&fix->param, &fix->status);
    g_assert_cmpint(ret, ==, 0);
    g_assert_cmpint(fix->param.target_id, ==, 1);
    g_assert(strlen(fix->param.targetname) > 0);
  }
}

static void test_get_data(TestFixture *fix, gconstpointer userData)
{
  int i, ret;

  for (i = 0; i < numCases; i++) {
    g_message("Scanning %s", cases[i].datafile);
    g_strlcpy(fix->param.datafile, cases[i].datafile, FLEN_FILENAME);
    fix->param.minband = cases[i].minband;
    fix->param.maxband = cases[i].maxband;
    ret = get_oi_fits_selection(&fix->param, &fix->status);
    g_assert_cmpint(ret, ==, 0);
    g_assert_cmpint(fix->param.target_id, ==, 1);
    g_assert(strlen(fix->param.targetname) > 0);

    g_message("Reading %s", cases[i].datafile);
    g_test_log_set_fatal_handler(ignoreRemoved, NULL);
    ret = get_oi_fits_data(&fix->param, &fix->data, &fix->status);
    g_assert_cmpint(ret, ==, 0);
    /* note BSMEM adds a power spectrum point at zero frequency: */
    g_assert_cmpint(fix->data.npow, ==, cases[i].numvis2 + 1);
    g_assert_cmpint(fix->data.nbis, ==, cases[i].numt3);
    g_assert_cmpint(fix->data.nuv, ==, cases[i].nuv);
    free_oi_data(&fix->data);
  }
}

int main(int argc, char *argv[])
{
  g_test_init(&argc, &argv, NULL);

  g_test_add("/bsmem/oifits/get_selection", TestFixture, NULL,
             setup_fixture, test_get_selection, teardown_fixture);
  g_test_add("/bsmem/oifits/get_data", TestFixture, NULL,
             setup_fixture, test_get_data, teardown_fixture);

  return g_test_run();
}
